#!/usr/bin/env node

'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var path = require('path');
var path__default = _interopDefault(path);
var rimraf = _interopDefault(require('rimraf'));
var webpack = _interopDefault(require('webpack'));
var series = _interopDefault(require('async/series'));
var gutil = _interopDefault(require('gulp-util'));
var fse = require('fs-extra');
var fse__default = _interopDefault(fse);
var fs = require('fs');
var AssetsPlugin = _interopDefault(require('assets-webpack-plugin'));
var builtinModules = _interopDefault(require('builtin-modules'));
var ExtractTextPlugin = _interopDefault(require('extract-text-webpack-plugin'));
var CodeSplitWebpackPlugin = _interopDefault(require('code-split-component/webpack'));
var HardSourceWebpackPlugin = _interopDefault(require('hard-source-webpack-plugin'));
var webpackBundleAnalyzer = require('webpack-bundle-analyzer');
var ChunkManifestPlugin = _interopDefault(require('chunk-manifest-webpack-plugin'));
var dotenv = _interopDefault(require('dotenv'));
var loaderUtils = require('loader-utils');
var postcssDevtools = _interopDefault(require('postcss-devtools'));
var atImport = _interopDefault(require('postcss-smart-import'));
var simpleUrl = _interopDefault(require('postcss-simple-url'));
var discardComments = _interopDefault(require('postcss-discard-comments'));
var advancedVariables = _interopDefault(require('postcss-advanced-variables'));
var sassyMixins = _interopDefault(require('postcss-sassy-mixins'));
var lost = _interopDefault(require('lost'));
var willChange = _interopDefault(require('postcss-will-change'));
var calc = _interopDefault(require('postcss-calc'));
var gradientTransparencyFix = _interopDefault(require('postcss-gradient-transparency-fix'));
var easings = _interopDefault(require('postcss-easings'));
var colorFunction = _interopDefault(require('postcss-color-function'));
var colorHexAlpha = _interopDefault(require('postcss-color-hex-alpha'));
var flexbugsFixes = _interopDefault(require('postcss-flexbugs-fixes'));
var mediaMinmax = _interopDefault(require('postcss-media-minmax'));
var customMedia = _interopDefault(require('postcss-custom-media'));
var nestedSelectors = _interopDefault(require('postcss-nested-ancestors'));
var nested = _interopDefault(require('postcss-nested'));
var gridKiss = _interopDefault(require('postcss-grid-kiss'));
var filters = _interopDefault(require('pleeease-filters'));
var transform = _interopDefault(require('postcss-transform-shortcut'));
var selectorMatches = _interopDefault(require('postcss-selector-matches'));
var pseudoelements = _interopDefault(require('postcss-pseudoelements'));
var autoprefixer = _interopDefault(require('autoprefixer'));
var reporter = _interopDefault(require('postcss-reporter'));
var hexrgba = _interopDefault(require('postcss-hexrgba'));
var unicodeChars = _interopDefault(require('postcss-unicode-characters'));
var zindex = _interopDefault(require('postcss-zindex'));
var warn = _interopDefault(require('postcss-at-warn'));
var responsiveType = _interopDefault(require('postcss-responsive-type'));
var inputStyles = _interopDefault(require('postcss-input-style'));
var assets = _interopDefault(require('postcss-assets'));
var lodash = require('lodash');
var chokidar = _interopDefault(require('chokidar'));
var express = _interopDefault(require('express'));
var createWebpackHotMiddleware = _interopDefault(require('webpack-hot-middleware'));
var createWebpackMiddleware = _interopDefault(require('webpack-dev-middleware'));
var notifier = _interopDefault(require('node-notifier'));

function compareModules(a,b) {
    if (a.resource < b.resource) {
        return -1;
    }
    if (a.resource > b.resource) {
        return 1;
    }
    return 0;
}

function getModuleSource (module) {
    var _source = module._source || {};
    return _source._value || "";
}

function concatenateSource (result, module_source) {
    return result + module_source;
}

function WebpackHashDigest () {

}

var hashType = "sha256";
var digestType = "base62";
var digestLength = 8;

WebpackHashDigest.prototype.apply = function(compiler) {
    compiler.plugin("compilation", function(compilation) {
        compilation.plugin("chunk-hash", function(chunk, chunkHash) {
            var source = chunk.modules.sort(compareModules).map(getModuleSource).reduce(concatenateSource, ''); // we provide an initialValue in case there is an empty module source. Ref: http://es5.github.io/#x15.4.4.21

            var generatedHash = loaderUtils.getHashDigest(source, hashType, digestType, digestLength);
            chunkHash.digest = function () {
                return generatedHash;
            };
        });
    });
};

var root = "node_modules";
var modules = new Set();

var nodePackages = fs.readdirSync(root).filter(function (dirname) { return dirname.charAt(0) !== "."; });

nodePackages.forEach(function (pkg) {
  try {
    var json = fse.readJsonSync(path.resolve(root, pkg, "package.json"));
  } catch (except) {
    return
  }

  if (json.module || json.style || json["jsnext:main"])
    { modules.add(pkg); }
});

var autoprefixerSettings = {
  browsers: "> 3%, last 2 versions, safari > 8, ie > 10, ios > 8, android > 4.3, samsung > 3, chromeandroid > 50",
  cascade: false
};

function getConfig(variables)
{
  if ( variables === void 0 ) variables = {};

  return [
    /*
    devtools({
      silent: true
    }),
    */

    atImport(),
    simpleUrl(),
    assets(),

    // Discard comments in your CSS files with PostCSS.
    // https://github.com/ben-eb/postcss-discard-comments
    // Remove all comments... we don't need them further down the line
    // which improves performance (reduces number of AST nodes)
    discardComments({
      removeAll: true
    }),

    // Automagical responsive typography. Adds a responsive property to font-size,
    // line-height and letter-spacing that generates complex calc and vw based font sizes.
    // https://github.com/seaneking/postcss-responsive-type
    responsiveType(),

    // Sass-like mixins
    // Needs to be executed before any variable handling plugins
    // https://github.com/andyjansson/postcss-sassy-mixins
    sassyMixins,

    // PostCSS plugin for Sass-like variables, conditionals, and iteratives
    // Supports local variables + @for/@each inspired by Sass
    // https://github.com/jonathantneal/postcss-advanced-variables
    advancedVariables({
      variables: variables
    }),

    // PostCSS plugin to reference any ancestor selector in nested CSS.
    // https://github.com/toomuchdesign/postcss-nested-ancestors
    nestedSelectors(),

    // Unwrap nested rules like how Sass does it.
    // https://github.com/postcss/postcss-nested
    nested,

    // Fractional grid system built with calc(). Supports masonry, vertical, and waffle grids.
    // https://github.com/peterramsing/lost
    lost,

    // A PostCSS plugin to keep CSS grids stupidly simple
    // See also: https://github.com/sylvainpolletvillard/postcss-grid-kiss
    gridKiss({
      fallback: true,
      screwIE: true
    }),

    // Insert 3D hack before will-change property
    // https://github.com/postcss/postcss-will-change
    willChange,

    // Reduce calc()
    // Note: Important to keep this after mixin/variable processing
    // https://github.com/postcss/postcss-calc
    calc,

    // Fix up CSS gradients with transparency for older browsers
    // https://github.com/gilmoreorless/postcss-gradient-transparency-fix
    gradientTransparencyFix,

    // Replace easing names from http://easings.net to `cubic-bezier()`.
    // https://github.com/postcss/postcss-easings
    easings,

    // Adds shorthand hex methods to rbga() values.
    // https://github.com/seaneking/postcss-hexrgba
    hexrgba,

    // Transform W3C CSS color function to more compatible CSS
    // https://github.com/postcss/postcss-color-function
    colorFunction,

    // Transform RGBA hexadecimal notations (#RRGGBBAA or #RGBA) to more compatible CSS (rgba())
    // https://github.com/postcss/postcss-color-hex-alpha
    colorHexAlpha,

    // Tries to fix all of flexbug's issues
    // https://github.com/luisrudge/postcss-flexbugs-fixes
    flexbugsFixes,

    // Writing simple and graceful Media Queries!
    // Support for CSS Media Queries Level 4: https://drafts.csswg.org/mediaqueries/#mq-range-context
    // https://github.com/postcss/postcss-media-minmax
    mediaMinmax,

    // Custom Media Queries
    // https://github.com/postcss/postcss-custom-media
    customMedia(),

    // Adds new pseudo-elements to inputs for easy cross-browser styling of their inner elements.
    // https://github.com/seaneking/postcss-input-style
    inputStyles,

    // Fallback for Webkit Filters property to SVG filters. Amazing stuff.
    // It converts all 10 CSS shorthand filters:
    // grayscale, sepia, saturate, hue-rotate, invert, opacity, brightness, contrast, blur, drop-shadow
    // https://github.com/iamvdo/pleeease-filters
    filters(),

    // Use shorthand transform properties in CSS
    // https://github.com/jonathantneal/postcss-transform-shortcut
    transform,

    // Transform :matches() W3C CSS pseudo class to more compatible CSS (simpler selectors)
    // https://github.com/postcss/postcss-selector-matches
    selectorMatches,

    // Add single and double colon peudo selectors
    // Normalizes e.g. `::before` to `:before` for wider browser support
    // https://github.com/axa-ch/postcss-pseudoelements
    pseudoelements,

    // An easier way to write unicode-range descriptors.
    // https://github.com/ben-eb/postcss-unicode-characters
    unicodeChars,

    // Reduce z-index values with PostCSS.
    // https://github.com/ben-eb/postcss-zindex
    zindex,

    // Sass like @warn for PostCSS. Disabling internal usage of "postcss-reporter".
    // https://github.com/ben-eb/postcss-at-warn
    warn({
      silent: true
    }),

    // Parse CSS and add vendor prefixes to rules by Can I Use
    // https://github.com/postcss/autoprefixer
    autoprefixer(autoprefixerSettings),

    // Log PostCSS messages to the console
    reporter({
      clearReportedMessages: true
    })
  ]
}

// import BabiliPlugin from "babili-webpack-plugin"
// Using more modern approach of hashing than "webpack-md5-hash". Somehow the SHA256 version
// ("webpack-sha-hash") does not correctly work based (produces different hashes for same content).
// This is basically a replacement of md5 with the loader-utils implementation which also supports
// shorter generated hashes based on base62 encoding instead of hex.
var builtInSet = new Set(builtinModules);

// - "intl" is included in one block with complete data. No reason for bundle everything here.
// - "react-intl" for the same reason as "intl" - contains a ton of locale data
// - "mime-db" database for working with mime types. Naturally pretty large stuff.
// - "helmet" uses some look with require which are not solvable with webpack
// - "express" also uses some dynamic requires
// - "encoding" uses dynamic iconv loading
// - "node-pre-gyp" native code module helper
// - "iltorb" brotli compression wrapper for NodeJS
// - "node-zopfli" native Zopfli implementation
var problematicCommonJS = new Set([ "intl", "react-intl", "mime-db", "helmet", "express", "encoding",
  "node-pre-gyp", "iltorb", "node-zopfli" ]);
var CWD = process.cwd();

// @see https://github.com/motdotla/dotenv
dotenv.config();



var VerboseProgressPlugin = function VerboseProgressPlugin(options) {
  this.options = options;
  this.cwd = process.cwd();
};

VerboseProgressPlugin.prototype.apply = function apply (compiler)
{
  compiler.plugin("compilation", function(compilation)
  {
    if (compilation.compiler.isChild())
      { return }

    var moduleCounter = 0;
    var activeModules = {};
    var slicePathBy = process.cwd().length + 1;

    function now(time) {
      return process.hrtime(time)
    }

    function moduleStart(module) {
      var ident = module.identifier();
      if (ident) {
        if (lodash.startsWith(ident, "ignored") || lodash.startsWith(ident, "external")) {
          return
        }
        moduleCounter++;
        if (moduleCounter % 100 === 0) {
          console.log("- Building module #" + moduleCounter);
        }
        activeModules[ident] = now();
      }
    }

    function moduleDone(module) {
      var ident = module.identifier();
      if (ident) {
        var entry = activeModules[ident];
        if (entry == null) {
          return
        }

        var runtime = Math.round(now(entry)[1] / 1000000);
        if (lodash.includes(ident, "!")) {
          var splits = ident.split("!");
          ident = splits.pop();
        }
        var relative = ident.slice(slicePathBy);
        // console.log(`Module ${relative} in ${runtime}ms`)
        if (splits) {
          splits = null;
        }
      }
    }

    compilation.plugin("build-module", moduleStart);
    compilation.plugin("failed-module", moduleDone);
    compilation.plugin("succeed-module", moduleDone);

    function log(title) {
      return function() {
        console.log("- " + title);
      }
    }

    compilation.plugin("seal", function() {
      console.log("- Sealing " + moduleCounter + " modules...");
    });
    compilation.plugin("optimize", log("Optimizing modules/chunks/tree..."));
    /*
    compilation.plugin("optimize-modules-basic", log("Basic module optimization"))
    compilation.plugin("optimize-modules", log("Module optimization"))
    compilation.plugin("optimize-modules-advanced", log("Advanced module optimization"))
    compilation.plugin("optimize-chunks-basic", log("Basic chunk optimization"))
    compilation.plugin("optimize-chunks", log("Chunk optimization"))
    compilation.plugin("optimize-chunks-advanced", log("Advanced chunk optimization"))
    compilation.plugin("optimize-tree", function(chunks, modules, callback) {
      console.log("- Module and chunk tree optimization")
      callback()
    })
    compilation.plugin("revive-modules", log("Module reviving"))
    compilation.plugin("optimize-module-order", log("Module order optimization"))
    compilation.plugin("optimize-module-ids", log("Module id optimization"))
    compilation.plugin("revive-chunks", log("Chunk reviving"))
    compilation.plugin("optimize-chunk-order", log("Chunk order optimization"))
    compilation.plugin("optimize-chunk-ids", log("Chunk id optimization"))
    compilation.plugin("before-hash", log("Hashing"))
    compilation.plugin("before-module-assets", log("Module assets processing"))
    compilation.plugin("before-chunk-assets", log("Chunk assets processing"))
    compilation.plugin("additional-chunk-assets", log("Additional chunk assets processing"))
    compilation.plugin("record", log("Recording"))
    compilation.plugin("additional-assets", function(callback) {
      console.log("- Additional asset processing")
      callback()
    })
    */
    compilation.plugin("optimize-chunk-assets", function(chunks, callback) {
      console.log("- Optimizing assets...");
      callback();
    });
    /*
    compilation.plugin("optimize-assets", function(assets, callback) {
      console.log("- Optimizing assets...")
      callback()
    })
    */
  });

  compiler.plugin("emit", function(compilation, callback) {
    console.log("- Emitting output files...");
    callback();
  });
};

function removeEmpty(array) {
  return array.filter(function (entry) { return Boolean(entry); })
}

function removeEmptyKeys(obj)
{
  var copy = {};
  for (var key in obj)
  {
    if (!(obj[key] == null || obj[key].length === 0))
      { copy[key] = obj[key]; }
  }

  return copy
}

function ifElse(condition) {
  return function (then, otherwise) { return (condition ? then : otherwise); }
}

function merge()
{
  var funcArgs = Array.prototype.slice.call(arguments); // eslint-disable-line prefer-rest-params

  return Object.assign.apply(
    null,
    removeEmpty([{}].concat(funcArgs))
  )
}

function isLoaderSpecificFile(request) {
  return Boolean(/\.(eot|woff|woff2|ttf|otf|svg|png|jpg|jpeg|gif|webp|webm|ico|mp4|mp3|ogg|html|pdf|swf|css|scss|sass|sss|less)$/.exec(request))
}

function ifIsFile(filePath) {
  try {
    return fs.statSync(filePath).isFile() ? filePath : ""
  } catch (err) {
    // empty
  }

  return ""
}


function getJsLoader(ref)
{
  var isNode = ref.isNode;
  var isWeb = ref.isWeb;
  var isProd = ref.isProd;
  var isDev = ref.isDev;

  var nodeBabel = isNode ? {
    // Don't try to find .babelrc because we want to force this configuration.
    babelrc: false,

    // Faster transpiling for minor loose in formatting
    compact: true,

    // Keep origin information alive
    sourceMaps: true,

    // Nobody needs the original comments when having source maps
    comments: false,

    presets:
    [
      // Exponentiation
      "babel-preset-es2016",

      // Async to generators + trailing function commas
      "babel-preset-es2017",

      // JSX, Flow
      "babel-preset-react"
    ],

    plugins:
    [
      // Transpile Markdown into React components. Super smart.
      "markdown-in-js/babel",

      // Optimization for lodash imports.
      // Auto cherry-picking es2015 imports from path imports.
      "lodash",

      // Keep transforming template literals as it keeps code smaller for the client
      // (removes multi line formatting which is allowed for literals)
      // This is interesting for all es2015 outputs... e.g.
      // later for a client build for modern builds, too
      "transform-es2015-template-literals",

      // class { handleClick = () => { } }
      // https://github.com/tc39/proposal-class-public-fields
      "transform-class-properties",

      // { ...todo, completed: true }
      [ "transform-object-rest-spread", { useBuiltIns: true }],

      // Polyfills the runtime needed
      [ "transform-runtime", { regenerator: false }],

      // Code Splitting by Routes
      [
        "code-split-component/babel", {
          disabled: isDev,
          mode: "server"
        }
      ]
    ]
  } : null;

  var webBabel = isWeb ? {
    // Don't try to find .babelrc because we want to force this configuration.
    babelrc: false,

    // Faster transpiling for minor loose in formatting
    compact: true,

    // Keep origin information alive
    sourceMaps: true,

    // Nobody needs the original comments when having source maps
    comments: false,

    presets:
    [
      // let, const, destructuring, classes, no modules
      [ "babel-preset-es2015", { modules: false }],

      // Exponentiation
      "babel-preset-es2016",

      // Async to generators + trailing function commas
      "babel-preset-es2017",

      // JSX, Flow
      "babel-preset-react"
    ],

    plugins:
    [
      // Transpile Markdown into React components. Super smart.
      "markdown-in-js/babel",

      // Optimization for lodash imports.
      // Auto cherry-picking es2015 imports from path imports.
      "lodash",

      // class { handleClick = () => { } }
      // https://github.com/tc39/proposal-class-public-fields
      "transform-class-properties",

      // { ...todo, completed: true }
      [ "transform-object-rest-spread", { useBuiltIns: true }],

      // Polyfills the runtime needed
      [ "transform-runtime", { regenerator: false }],

      // Code Splitting by Routes
      [
        "code-split-component/babel", {
          disabled: isDev,
          mode: "client"
        }
      ]
    ]
  } : null;

  return [{
    loader: "babel-loader",
    options: merge(
      {
        // Enable caching for babel transpiles
        cacheDirectory: true,

        env:
        {
          production: {
            comments: false,
            plugins: [
              // Cleanup descriptions for translations from compilation output
              "react-intl",

              // Remove prop types from our code
              "transform-react-remove-prop-types"
            ]
          },

          development: {
            plugins: [
              // Adds component stack to warning messages
              "transform-react-jsx-source",

              // Adds __self attribute to JSX which React will use for some warnings
              "transform-react-jsx-self",

              // Add deprecation messages
              "log-deprecated",

              // React based
              "react-hot-loader/babel"
            ]
          }
        }
      },

      nodeBabel,
      webBabel
    )
  }]
}


function getCssLoaders(ref)
{
  var isNode = ref.isNode;
  var isWeb = ref.isWeb;
  var isProd = ref.isProd;
  var isDev = ref.isDev;

  // When targetting the node we fake out the style loader as the
  // node can't handle the styles and doesn't care about them either..
  if (isNode)
  {
    return [
      {
        loader: "css-loader/locals",
        query:
        {
          sourceMap: false,
          modules: true,
          localIdentName: isProd ? "[local]-[hash:base62:8]" : "[path][name]-[local]"
        }
      },
      {
        loader: "postcss-loader"
      }
    ]
  }

  if (isWeb)
  {
    // For a production client build we use the ExtractTextPlugin which
    // will extract our CSS into CSS files. The plugin needs to be
    // registered within the plugins section too.
    if (isProd)
    {
      // First: the loader(s) that should be used when the css is not extracted
      // Second: the loader(s) that should be used for converting the resource to a css exporting module
      // Note: Unfortunately it seems like it does not support the new query syntax of webpack v2
      // See also: https://github.com/webpack/extract-text-webpack-plugin/issues/196
      return ExtractTextPlugin.extract({
        allChunks: true,
        fallbackLoader: "style-loader",
        loader:
        [
          {
            loader: "css-loader",
            query:
            {
              sourceMap: true,
              modules: true,
              localIdentName: "[local]-[hash:base62:8]",
              minimize: false,
              import: false
            }
          },
          {
            loader: "postcss-loader"
          }
        ]
      })
    }
    else
    {
      // For a development client we will use a straight style & css loader
      // along with source maps. This combo gives us a better development
      // experience.
      return [
        {
          loader: "style-loader"
        },
        {
          loader: "css-loader",
          query:
          {
            sourceMap: true,
            modules: true,
            localIdentName: "[path][name]-[local]",
            minimize: false,
            import: false
          }
        },
        {
          loader: "postcss-loader"
        }
      ]
    }
  }
}

var isDebug = true;
var isVerbose = true;

function ConfigFactory(target, mode, options, root)
{
  if ( options === void 0 ) options = {};
  if ( root === void 0 ) root = CWD;

  // Output custom options
  if (Object.keys(options).length > 0) {
    console.log("Using options: ", options);
  }

  if (!target || !~[ "web", "node" ].findIndex(function (valid) { return target === valid; }))
  {
    throw new Error(
      "You must provide a \"target\" (web|node) to the ConfigFactory."
    )
  }

  if (!mode || !~[ "development", "production" ].findIndex(function (valid) { return mode === valid; }))
  {
    throw new Error(
      "You must provide a \"mode\" (development|production) to the ConfigFactory."
    )
  }

  process.env.NODE_ENV = options.debug ? "development" : mode;
  process.env.BABEL_ENV = mode;

  var isDev = mode === "development";
  var isProd = mode === "production";
  var isWeb = target === "web";
  var isNode = target === "node";

  var ifDev = ifElse(isDev);
  var ifProd = ifElse(isProd);
  var ifWeb = ifElse(isWeb);
  var ifNode = ifElse(isNode);
  var ifDevWeb = ifElse(isDev && isWeb);
  var ifDevNode = ifElse(isDev && isNode);
  var ifProdWeb = ifElse(isProd && isWeb);
  var ifProdNode = ifElse(isProd && isNode);
  var ifIntegration = ifElse(process.env.CI || false);
  var ifUniversal = ifElse(process.env.DISABLE_SSR);


  var folder = isWeb ? "client" : "server";

  var cssLoaders = getCssLoaders({
    isProd: isProd,
    isDev: isDev,
    isWeb: isWeb,
    isNode: isNode
  });

  var jsLoaders = getJsLoader({
    isProd: isProd,
    isDev: isDev,
    isWeb: isWeb,
    isNode: isNode
  });

  var excludeFromTranspilation = [
    /node_modules/,
    path__default.resolve(root, process.env.CLIENT_BUNDLE_OUTPUT_PATH),
    path__default.resolve(root, process.env.SERVER_BUNDLE_OUTPUT_PATH)
  ];

  // Just bundle the NodeJS files which are from the local project instead
  // of a deep self-contained bundle.
  // See also: https://nolanlawson.com/2016/08/15/the-cost-of-small-modules/
  var useLightNodeBundle = options.lightBundle == null ? isDev : options.lightBundle;
  if (useLightNodeBundle && isNode) {
    console.log("Using light node bundle");
  }

  return {
    // We need to inform Webpack about our build target
    target: target,

    // We have to set this to be able to use these items when executing a
    // node bundle. Otherwise strangeness happens, like __dirname resolving
    // to '/'. There is no effect on our client bundle.
    node: {
      __dirname: true,
      __filename: true
    },

    // What information should be printed to the console
    stats: {
      colors: true,
      reasons: isDebug,
      hash: isVerbose,
      version: isVerbose,
      timings: true,
      chunks: isVerbose,
      chunkModules: isVerbose,
      cached: isVerbose,
      cachedAssets: isVerbose
    },

    // This is not the file cache, but the runtime cache.
    // The reference is used to speed-up rebuilds in one execution e.g. via watcher
    // Note: But is has to share the same configuration as the cache is not config aware.
    // cache: cache,

    // Capture timing information for each module.
    // Analyse tool: http://webpack.github.io/analyse
    profile: isProd,

    // Report the first error as a hard error instead of tolerating it.
    bail: isProd,

    // Anything listed in externals will not be included in our bundle.
    externals: removeEmpty([
      ifNode(function(context, request, callback)
      {
        var basename = request.split("/")[0];

        // Externalize built-in modules
        if (builtInSet.has(basename))
          { return callback(null, "commonjs " + request) }

        // Keep care that problematic common-js code is external
        if (problematicCommonJS.has(basename))
          { return callback(null, "commonjs " + request) }

        // Ignore inline files
        if (basename.charAt(0) === ".")
          { return callback() }

        // But inline all es2015 modules
        if (modules.has(basename))
          { return callback() }

        // Inline all files which are dependend on Webpack loaders e.g. CSS, images, etc.
        if (isLoaderSpecificFile(request))
          { return callback() }

        // In all other cases follow the user given preference
        useLightNodeBundle ? callback(null, "commonjs " + request) : callback();
      })
    ]),

    // See also: https://webpack.github.io/docs/configuration.html#devtool
    // and http://webpack.github.io/docs/build-performance.html#sourcemaps
    // All "module*" and "cheap" variants do not seem to work with this kind
    // of setup where we have loaders involved. Even simple console messages jump
    // to the wrong location in these cases.
    devtool: "source-map",

    // New performance hints. Only active for production build.
    performance: {
      hints: isProd && isWeb ? "warning" : false
    },

    // Define our entry chunks for our bundle.
    entry: removeEmptyKeys(
    {
      main: removeEmpty([
        ifDevWeb("react-hot-loader/patch"),
        ifDevWeb(("webpack-hot-middleware/client?reload=true&path=http://localhost:" + (process.env.CLIENT_DEVSERVER_PORT) + "/__webpack_hmr")),
        options.entry ? options.entry : ifIsFile(("./src/" + folder + "/index.js")) ]),

      vendor: ifProdWeb(options.vendor ? options.vendor : ifIsFile(("./src/" + folder + "/vendor.js")))
    }),

    output:
    {
      // The dir in which our bundle should be output.
      path: path__default.resolve(
        root,
        isWeb ? process.env.CLIENT_BUNDLE_OUTPUT_PATH : process.env.SERVER_BUNDLE_OUTPUT_PATH
      ),

      // The filename format for our bundle's entries.
      filename: ifProdWeb(

        // We include a hash for client caching purposes. Including a unique
        // has for every build will ensure browsers always fetch our newest
        // bundle.
        "[name]-[chunkhash].js",

        // We want a determinable file name when running our NodeJS bundles,
        // as we need to be able to target our NodeJS start file from our
        // npm scripts. We don't care about caching on the NodeJS anyway.
        // We also want our client development builds to have a determinable
        // name for our hot reloading client bundle NodeJS.
        "[name].js"
      ),

      chunkFilename: ifProdWeb(
        "chunk-[name]-[chunkhash].js",
        "chunk-[name].js"
      ),

      // This is the web path under which our webpack bundled output should
      // be considered as being served from.
      publicPath: ifDev(

        // As we run a seperate NodeJS for our client and NodeJS bundles we
        // need to use an absolute http path for our assets public path.
        ("http://localhost:" + (process.env.CLIENT_DEVSERVER_PORT) + (process.env.CLIENT_BUNDLE_HTTP_PATH)),

        // Otherwise we expect our bundled output to be served from this path.
        process.env.CLIENT_BUNDLE_HTTP_PATH
      ),

      // When in NodeJS mode we will output our bundle as a commonjs2 module.
      libraryTarget: ifNode("commonjs2", "var")
    },

    resolve:
    {
      // Enable new module/jsnext:main field for requiring files
      // Defaults: https://webpack.github.io/docs/configuration.html#resolve-packagemains
      mainFields: ifNode(
        [ "module", "jsnext:main", "main" ],
        [ "web", "browser", "style", "module", "jsnext:main", "main" ]
      ),

      // These extensions are tried when resolving a file.
      extensions: [
        ".js",
        ".jsx",
        ".ts",
        ".tsx",
        ".css",
        ".sss",
        ".json"
      ]
    },

    plugins: removeEmpty([
      // Improve source caching in Webpack v2.
      // Therefor we disable it in production and only use it to speed up development rebuilds.
      ifDev(new HardSourceWebpackPlugin({
        // Either an absolute path or relative to output.path.
        cacheDirectory: path__default.resolve(root, ".hardsource", (target + "-" + mode)),

        // Either an absolute path or relative to output.path. Sets webpack's
        // recordsPath if not already set.
        recordsPath: path__default.resolve(root, ".hardsource", (target + "-" + mode), "records.json"),

        // Optional field. This field determines when to throw away the whole
        // cache if for example npm modules were updated.
        environmentHash: {
          root: root,
          directories: [ "node_modules" ],
          files: [ "package.json", "yarn.lock" ]
        }
      })),

      // Adds options to all of our loaders.
      ifDev(
        new webpack.LoaderOptionsPlugin({
          // Indicates to our loaders that they should minify their output
          // if they have the capability to do so.
          minimize: false,

          // Indicates to our loaders that they should enter into debug mode
          // should they support it.
          debug: true,

          // Pass options for PostCSS
          options: {
            postcss: getConfig({}),
            context: CWD
          }
        })
      ),

      // Adds options to all of our loaders.
      ifProd(
        new webpack.LoaderOptionsPlugin({
          // Indicates to our loaders that they should minify their output
          // if they have the capability to do so.
          minimize: true,

          // Indicates to our loaders that they should enter into debug mode
          // should they support it.
          debug: false,

          // Pass options for PostCSS
          options: {
            postcss: getConfig({}),
            context: CWD
          }
        })
      ),

      new VerboseProgressPlugin(),

      /*
      ifProd(new BabiliPlugin({
        comments: false,
        babili: {
          minified: true,
          comments: false,
          plugins: [
            "minify-dead-code-elimination",
            "minify-mangle-names",
            "minify-simplify"
          ]
        }
      })),
      */

      new CodeSplitWebpackPlugin({
        // The code-split-component doesn't work nicely with hot module reloading,
        // which we use in our development builds, so we will disable it if we
        // are creating a development bundle (which results in synchronous loading
        // behavior on the CodeSplit instances).
        disabled: isDev
      }),

      // Proposed fix to work around issues with CodeSplit vs. HardSource plugin config
      // Via: https://github.com/sebastian-software/advanced-boilerplate/compare/master...mzgoddard:code-split-records-compat
      {
        apply: function(compiler) {
          var safeChunkIdMax = 100000;
          compiler.plugin("compilation", function(compilation) {
            compilation.plugin("optimize-chunk-order", function() {
              if (compilation.usedChunkIds) {
                var usedChunkIds = {};
                for (var key in compilation.usedChunkIds) {
                  if (compilation.usedChunkIds[key] < safeChunkIdMax) {
                    usedChunkIds[key] = compilation.usedChunkIds[key];
                  }
                }
                compilation.usedChunkIds = usedChunkIds;
              }
            });
          });
        }
      },

      // For NodeJS bundle, you also want to use "source-map-support" which automatically sourcemaps
      // stack traces from NodeJS. We need to install it at the top of the generated file, and we
      // can use the BannerPlugin to do this.
      // - `raw`: true tells webpack to prepend the text as it is, not wrapping it in a comment.
      // - `entryOnly`: false adds the text to all generated files, which you might have multiple if using code splitting.
      // Via: http://jlongster.com/Backend-Apps-with-Webpack--Part-I
      ifNode(new webpack.BannerPlugin({
        banner: 'require("source-map-support").install();',
        raw: true,
        entryOnly: false
      })),

      // Extract vendor bundle for keeping larger parts of the application code
      // delivered to users stable during development (improves positive cache hits)
      ifProdWeb(new webpack.optimize.CommonsChunkPlugin({
        name: "vendor",
        minChunks: Infinity
      })),

      // More aggressive chunk merging strategy. Even similar chunks are merged if the
      // total size is reduced enough.
      ifProdWeb(new webpack.optimize.AggressiveMergingPlugin()),

      // We use this so that our generated [chunkhash]'s are only different if
      // the content for our respective chunks have changed.  This optimises
      // our long term browser caching strategy for our client bundle, avoiding
      // cases where browsers end up having to download all the client chunks
      // even though 1 or 2 may have only changed.
      ifProdWeb(new WebpackHashDigest()),

      // Extract chunk hashes into separate file
      ifProdWeb(new ChunkManifestPlugin({
        filename: "manifest.json",
        manifestVariable: "CHUNK_MANIFEST"
      })),

      // Each key passed into DefinePlugin is an identifier.
      // The values for each key will be inlined into the code replacing any
      // instances of the keys that are found.
      // If the value is a string it will be used as a code fragment.
      // If the value isn’t a string, it will be stringified (including functions).
      // If the value is an object all keys are removeEmpty the same way.
      // If you prefix typeof to the key, it’s only removeEmpty for typeof calls.
      new webpack.DefinePlugin({
        "process.env.TARGET": JSON.stringify(target),
        "process.env.MODE": JSON.stringify(mode),

        // NOTE: The NODE_ENV key is especially important for production
        // builds as React relies on process.env.NODE_ENV for optimizations.
        "process.env.NODE_ENV": JSON.stringify(process.env.NODE_ENV),

        "process.env.APP_ROOT": JSON.stringify(path__default.resolve(root)),

        // All the below items match the config items in our .env file. Go
        // to the .env.example for a description of each key.
        "process.env.SERVER_PORT": process.env.SERVER_PORT,
        "process.env.CLIENT_DEVSERVER_PORT": process.env.CLIENT_DEVSERVER_PORT,

        "process.env.DISABLE_SSR": process.env.DISABLE_SSR,

        "process.env.SERVER_BUNDLE_OUTPUT_PATH": JSON.stringify(process.env.SERVER_BUNDLE_OUTPUT_PATH),
        "process.env.CLIENT_BUNDLE_OUTPUT_PATH": JSON.stringify(process.env.CLIENT_BUNDLE_OUTPUT_PATH),
        "process.env.CLIENT_PUBLIC_PATH": JSON.stringify(process.env.CLIENT_PUBLIC_PATH),

        "process.env.CLIENT_BUNDLE_ASSETS_FILENAME": JSON.stringify(process.env.CLIENT_BUNDLE_ASSETS_FILENAME),
        "process.env.CLIENT_BUNDLE_CHUNK_MANIFEST_FILENAME": JSON.stringify(process.env.CLIENT_BUNDLE_CHUNK_MANIFEST_FILENAME),

        "process.env.CLIENT_BUNDLE_HTTP_PATH": JSON.stringify(process.env.CLIENT_BUNDLE_HTTP_PATH),
        "process.env.CLIENT_BUNDLE_CACHE_MAXAGE": JSON.stringify(process.env.CLIENT_BUNDLE_CACHE_MAXAGE)
      }),

      // Generates a JSON file containing a map of all the output files for
      // our webpack bundle. A necessisty for our NodeJS rendering process
      // as we need to interogate these files in order to know what JS/CSS
      // we need to inject into our HTML.
      ifWeb(
        new AssetsPlugin({
          filename: process.env.CLIENT_BUNDLE_ASSETS_FILENAME,
          path: path__default.resolve(root, process.env.CLIENT_BUNDLE_OUTPUT_PATH),
          prettyPrint: true
        })
      ),

      // Effectively fake all "file-loader" files with placeholders on NodeJS
      ifNode(new webpack.NormalModuleReplacementPlugin(
        /\.(eot|woff|woff2|ttf|otf|svg|png|jpg|jpeg|gif|webp|webm|mp4|mp3|ogg|html|pdf)$/,
        "node-noop"
      )),

      // We don't want webpack errors to occur during development as it will
      // kill our development NodeJS instances.
      ifDev(new webpack.NoEmitOnErrorsPlugin()),

      // We need this plugin to enable hot module reloading for our dev NodeJS.
      ifDevWeb(new webpack.HotModuleReplacementPlugin()),

      // This is a production client so we will extract our CSS into
      // CSS files.
      ifProdWeb(
        new ExtractTextPlugin({
          filename: "[name]-[contenthash:base62:8].css",

          // FIXME: Work on some alternative approach for ExtractText to produce multiple
          // external CSS files.
          //
          // Possibly fork: https://github.com/webpack/extract-text-webpack-plugin
          // The current allChunks flag extract all styles from all chunks into one CSS file.
          // Without this flag all non entry chunks are keeping their CSS inline in the JS bundle.
          //
          // Idea: Split into multiple chunk oriented CSS files + load the CSS files during
          // from inside the JS file e.g. via calling some external API to load stylesheets.
          allChunks: true
        })
      ),

      // Analyse webpack bundle
      ifProdWeb(new webpackBundleAnalyzer.BundleAnalyzerPlugin({
        openAnalyzer: false,
        analyzerMode: "static"
      }))
    ]),

    module:
    {
      rules: removeEmpty(
      [
        // JavaScript
        {
          test: /\.(js|jsx)$/,
          loaders: jsLoaders,
          exclude: excludeFromTranspilation
        },

        // Typescript
        // https://github.com/s-panferov/awesome-typescript-loader
        {
          test: /\.(ts|tsx)$/,
          loader: "awesome-typescript-loader",
          exclude: excludeFromTranspilation
        },

        // CSS
        {
          test: /\.css$/,
          loader: cssLoaders
        },

        // JSON
        {
          test: /\.json$/,
          loader: "json-loader"
        },

        // YAML
        {
          test: /\.(yml|yaml)$/,
          loaders: [ "json-loader", "yaml-loader" ]
        },

        // References to images, fonts, movies, music, etc.
        {
          test: /\.(eot|woff|woff2|ttf|otf|svg|png|jpg|jpeg|jp2|jpx|jxr|gif|webp|mp4|mp3|ogg|pdf|html)$/,
          loader: "file-loader",
          options: {
            name: ifProdWeb("file-[hash:base62:8].[ext]", "[name].[ext]"),
            emitFile: isWeb
          }
        },

        // GraphQL support
        // @see http://dev.apollodata.com/react/webpack.html
        {
          test: /\.(graphql|gql)$/,
          loader: "graphql-tag/loader",
          exclude: excludeFromTranspilation
        }
      ])
    }
  }
}

function build()
{
  var buildFolderClient = path__default.resolve("build", "client");
  var buildFolderServer = path__default.resolve("build", "server");

  series([
    function(callback)
    {
      // Remove all content but keep the directory so that
      // if you're in it, you don't end up in Trash
      rimraf(buildFolderClient + "/*", callback);
    },
    function(callback)
    {
      // Remove all content but keep the directory so that
      // if you're in it, you don't end up in Trash
      rimraf(buildFolderServer + "/*", callback);
    },
    function(callback)
    {
      console.log("Creating a production build for client...");
      webpack(ConfigFactory("web", "production")).run(function (err, stats) {
        if (err)
        {
          console.error("Failed to create a production build for client:");
          console.error(err.message || err);
          process.exit(1);
        }

        console.log("- Done");
        console.log("");

        gutil.log(stats.toString({
          children: false,
          chunks: false,
          colors: true
        }));

        //fixme: remove this snippet when https://github.com/webpack/webpack/issues/2390 is fixed
        if (stats.hasErrors()) {
          process.exit(1);
        }

        var jsonStats = stats.toJson();
        fse__default.writeJsonSync(path__default.resolve(buildFolderClient, "stats.json"), jsonStats);
        callback();
      });
    },
    function(callback)
    {
      console.log("Creating a production build for node...");
      webpack(ConfigFactory("node", "production")).run(function (err, stats) {
        if (err)
        {
          console.error("Failed to create a production build for server:");
          console.error(err.message || err);
          process.exit(1);
        }

        console.log("- Done");
        console.log("");

        gutil.log(stats.toString({
          children: false,
          chunks: false,
          colors: true
        }));

        //fixme: remove this snippet when https://github.com/webpack/webpack/issues/2390 is fixed
        if (stats.hasErrors()) {
          process.exit(1);
        }

        var jsonStats = stats.toJson();
        fse__default.writeJsonSync(path__default.resolve(buildFolderServer, "stats.json"), jsonStats);
        callback();
      });
    }
  ]);
}

var ListenerManager = function ListenerManager(server) {
  var this$1 = this;

  this.lastConnectionKey = 0;
  this.connectionMap = {};
  this.started = false;

  server.then(function (listener) {
    this$1.started = true;
    this$1.listener = listener;

    // Track all connections to our server so that we can close them when needed.
    this$1.listener.on("connection", function (connection) {
      // Generate a new key to represent the connection
      var connectionKey = ++this$1.lastConnectionKey;

      // Add the connection to our map.
      this$1.connectionMap[connectionKey] = connection;

      // Remove the connection from our map when it closes.
      connection.on("close", function () {
        delete this$1.connectionMap[connectionKey];
      });
    });
  });
};

ListenerManager.prototype.killAllConnections = function killAllConnections () {
    var this$1 = this;

  Object.keys(this.connectionMap).forEach(function (connectionKey) {
    this$1.connectionMap[connectionKey].destroy();
  });
};

ListenerManager.prototype._tryDispose = function _tryDispose (force) {
    var this$1 = this;
    if ( force === void 0 ) force = false;

  return new Promise(function (resolve$$1, reject) {
    if (force) {
      // Forcefully close any existing connections.
      this$1.killAllConnections();
    }

    // Close the listener.
    if (this$1.listener) {
      this$1.listener.close(function (err) {
        if (err) {
          // reject when server is not open when it is closed
          // see https://nodejs.org/api/net.html#net_server_close_callback
          reject(err);
        }
        // Ensure no straggling connections are left over.
        this$1.killAllConnections();

        resolve$$1();
      });
    } else {
      resolve$$1();
    }
  })
};

ListenerManager.prototype.dispose = function dispose (force) {
    var this$1 = this;
    if ( force === void 0 ) force = false;

  return this._tryDispose(force)
    .catch(
      // postpone 100ms to retry
      function () { return new Promise(function (resolve$$1) { return setTimeout(resolve$$1, 100); }).then(function () { return this$1.dispose(force); }); }
    )
};

function createNotification(options)
{
  if ( options === void 0 ) options = {};

  notifier.notify({
    title: options.title,
    message: options.message,
    open: options.open
  });

  console.log(((options.title) + ": " + (options.message)));
}

var HotClient = function HotClient(compiler) {
  var this$1 = this;

  this.listenerManager = new ListenerManager(new Promise(function (resolve$$1, reject) {

    var app = express();
    this$1.webpackDevMiddleware = createWebpackMiddleware(compiler, {
      noInfo: true,
      headers: {
        "Access-Control-Allow-Origin": "*"
      },

      // The path at which the client bundles are served from.Note: in this
      // case as we are running a seperate dev server the public path should
      // be absolute, i.e. including the "http://..."
      publicPath: compiler.options.output.publicPath,

      // Make it beautiful
      stats: {
        colors: true
      }
    });
    app.use(this$1.webpackDevMiddleware);
    app.use(createWebpackHotMiddleware(compiler));

    var listener = app.listen(process.env.CLIENT_DEVSERVER_PORT);
    createNotification({
      title: "Hot Development Client",
      message: ("Started successfully (Port: " + (process.env.CLIENT_DEVSERVER_PORT) + ")!")
    });

    resolve$$1(listener);
  }));
};

HotClient.prototype.dispose = function dispose (force) {
    if ( force === void 0 ) force = false;

  this.webpackDevMiddleware.close();

  return this.listenerManager ? this.listenerManager.dispose(force) : Promise.resolve()
};

var HotServer = function HotServer(compiler) {
  this.compiler = compiler;
  this.listenerManager = null;

  var compiledOutputPath = path.resolve(
    compiler.options.output.path, ((Object.keys(compiler.options.entry)[0]) + ".js")
  );

  var importedServer;
  try
  {
    importedServer = require(compiledOutputPath);
  }
  catch (err)
  {
    createNotification({
      title: "Hot Server",
      message: "Error: Bundle invalid. Check console for error!"
    });

    console.error(err);
  }

  if (importedServer) {
    var listener = importedServer.start();
    createNotification({
      title: "Hot Server",
      message: "Started successfully!"
    });

    this.listenerManager = new ListenerManager(listener);
  }
};

HotServer.prototype.dispose = function dispose (force) {
    if ( force === void 0 ) force = false;

  return this.listenerManager ? this.listenerManager.dispose(force) : Promise.resolve()
};

var CWD$1 = process.cwd();

var HotServers = function HotServers(root)
{
  var this$1 = this;
  if ( root === void 0 ) root = CWD$1;

  this.root = root;

  // Bind our functions to avoid any scope/closure issues.
  this.start = this.start.bind(this);
  this.restart = this.restart.bind(this);

  this._configureHotClient = this._configureHotClient.bind(this);
  this._configureHotServer = this._configureHotServer.bind(this);

  this.clientBundle = null;
  this.clientCompiler = null;
  this.serverBundle = null;
  this.serverCompiler = null;

  // If we receive a kill cmd then we will first try to dispose our listeners.
  process.on("SIGTERM", function () { return this$1.dispose().then(function () { return process.exit(0); }); });
};

HotServers.prototype.start = function start ()
{
  try
  {
    console.log("Compiling client...");
    this.clientCompiler = webpack(ConfigFactory("web", "development"));

    console.log("Compiling server...");
    this.serverCompiler = webpack(ConfigFactory("node", "development"));
  }
  catch (err) {
    createNotification({
      title: "Webpack",
      message: "Config invalid, check console for error"
    });
    console.error(err);
    return
  }

  this._configureHotClient();
  this._configureHotServer();
};


HotServers.prototype.dispose = function dispose ()
{
    var this$1 = this;

  // We want to forcefully close our servers (passing true) which will hard
  // kill any existing connections.We don't care about them running as we
  // need to restart both the client and server bundles.
  var safeDisposeClient = function () { return (this$1.clientBundle ? this$1.clientBundle.dispose(true) : Promise.resolve()); };
  var safeDisposeServer = function () { return (this$1.serverBundle ? this$1.serverBundle.dispose(true) : Promise.resolve()); };

  return safeDisposeClient().then(safeDisposeServer)
};

HotServers.prototype.restart = function restart () {
  var clearWebpackConfigsCache = function () {
    Object.keys(require.cache).forEach(function (modulePath) {
      if (~modulePath.indexOf("webpack")) {
        delete require.cache[modulePath];
      }
    });
  };

  this.dispose()
    .then(clearWebpackConfigsCache)
    .then(this.start, function (err) { return console.log(err); })
    .catch(function (err) { return console.log(err); });
};

HotServers.prototype._configureHotClient = function _configureHotClient () {
  this.clientCompiler.plugin("done", function (stats) {
    if (stats.hasErrors()) {
      createNotification({
        title: "Client",
        message: "Build failed, check console for error"
      });
      console.log(stats.toString());
    }
  });

  this.clientBundle = new HotClient(this.clientCompiler);
};

HotServers.prototype._configureHotServer = function _configureHotServer () {
    var this$1 = this;

  var compileHotServer = function () {
    var runCompiler = function () { return this$1.serverCompiler.run(function () { return undefined; }); };

    // Shut down any existing running server if necessary before starting the
    // compile, else just compile.
    if (this$1.serverBundle) {
      this$1.serverBundle.dispose(true).then(runCompiler);
    } else {
      runCompiler();
    }
  };

  this.clientCompiler.plugin("done", function (stats) {
    if (!stats.hasErrors()) {
      compileHotServer();
    }
  });

  this.serverCompiler.plugin("done", function (stats) {
    if (stats.hasErrors()) {
      createNotification({
        title: "Server",
        message: "Build failed, check console for error"
      });
      console.log(stats.toString());
      return
    }

    // Make sure our newly built server bundles aren't in the module cache.
    Object.keys(require.cache).forEach(function (modulePath) {
      if (~modulePath.indexOf(this$1.serverCompiler.options.output.path)) {
        delete require.cache[modulePath];
      }
    });

    this$1.serverBundle = new HotServer(this$1.serverCompiler);
  });

  // Now we will configure `chokidar` to watch our server specific source folder.
  // Any changes will cause a rebuild of the server bundle.
  this.watcher = chokidar.watch([ path.resolve(this.root, "./src/server") ]);
  this.watcher.on("ready", function () {
    this$1.watcher
      .on("add", compileHotServer)
      .on("addDir", compileHotServer)
      .on("change", compileHotServer)
      .on("unlink", compileHotServer)
      .on("unlinkDir", compileHotServer);
  });
};

function start() {
  new HotServers().start();
}

var script = process.argv[2];
var args = process.argv.slice(3);

switch (script)
{
  case "build":
    build(args);
    break

  case "start":
    start(args);
    break

  case null:
  case undefined:
    console.log("No script name given!");
    break

  default:
    console.log(("Unknown script \"" + script + "\"!"));
    break
}
//# sourceMappingURL=advanced-script.js.map
